@@ -55,9 +55,11 @@ module Data.ASN1.Prim
 	) where
 
 import Data.ASN1.Internal
-import Data.ASN1.Raw
 import Data.ASN1.Stream
 import Data.ASN1.BitArray
+import Data.ASN1.Types
+import Data.ASN1.Serialize
+import Data.Serialize.Put (runPut)
 import Data.Bits
 import Data.Word
 import Data.List (unfoldr)
@@ -69,6 +71,8 @@ import Data.Text.Lazy (Text)
 import qualified Data.Text.Lazy as T
 import Data.Text.Lazy.Encoding (decodeASCII, decodeUtf8, decodeUtf32BE, encodeUtf8, encodeUtf32BE)
 
+import Control.Applicative
+
 encodeUCS2BE :: Text -> L.ByteString
 encodeUCS2BE t =
 	L.pack $ concatMap (\c -> let (d,m) = (fromEnum c) `divMod` 256 in [fromIntegral d,fromIntegral m] ) $ T.unpack t
@@ -152,7 +156,7 @@ encodePrimitive a =
 	let blen = B.length b in
 	let len = makeLength blen in
 	let hdr = encodePrimitiveHeader len a in
-	(B.length (putHeader hdr) + blen, [Header hdr, Primitive b])
+	(B.length (runPut $ putHeader hdr) + blen, [Header hdr, Primitive b])
 	where
 		makeLength len
 			| len < 0x80 = LenShort len
@@ -182,7 +186,7 @@ encodeConstructed c@(Start _) children =
 	let (clen, events) = encodeList children in
 	let len = mkSmallestLength clen in
 	let h = encodeHeader True len c in
-	let tlen = B.length (putHeader h) + clen in
+	let tlen = B.length (runPut $ putHeader h) + clen in
 	(tlen, Header h : ConstructionBegin : events ++ [ConstructionEnd])
 
 encodeConstructed _ _ = error "not a start node"
@@ -193,7 +197,7 @@ mkSmallestLength i
 	| otherwise = LenLong (nbBytes i) i
 		where nbBytes nb = if nb > 255 then 1 + nbBytes (nb `div` 256) else 1
 
-type ASN1Ret = Either ASN1Err ASN1
+type ASN1Ret = Either ASN1Error ASN1
 
 decodePrimitive :: ASN1Header -> B.ByteString -> ASN1Ret
 decodePrimitive (ASN1Header Universal 0x1 _ _) p   = getBoolean False p
@@ -202,13 +206,13 @@ decodePrimitive (ASN1Header Universal 0x3 _ _) p   = getBitString p
 decodePrimitive (ASN1Header Universal 0x4 _ _) p   = getOctetString p
 decodePrimitive (ASN1Header Universal 0x5 _ _) p   = getNull p
 decodePrimitive (ASN1Header Universal 0x6 _ _) p   = getOID p
-decodePrimitive (ASN1Header Universal 0x7 _ _) _   = Left $ ASN1NotImplemented "Object Descriptor"
-decodePrimitive (ASN1Header Universal 0x8 _ _) _   = Left $ ASN1NotImplemented "External"
-decodePrimitive (ASN1Header Universal 0x9 _ _) _   = Left $ ASN1NotImplemented "real"
-decodePrimitive (ASN1Header Universal 0xa _ _) _   = Left $ ASN1NotImplemented "enumerated"
-decodePrimitive (ASN1Header Universal 0xb _ _) _   = Left $ ASN1NotImplemented "EMBEDDED PDV"
+decodePrimitive (ASN1Header Universal 0x7 _ _) _   = Left $ TypeNotImplemented "Object Descriptor"
+decodePrimitive (ASN1Header Universal 0x8 _ _) _   = Left $ TypeNotImplemented "External"
+decodePrimitive (ASN1Header Universal 0x9 _ _) _   = Left $ TypeNotImplemented "real"
+decodePrimitive (ASN1Header Universal 0xa _ _) _   = Left $ TypeNotImplemented "enumerated"
+decodePrimitive (ASN1Header Universal 0xb _ _) _   = Left $ TypeNotImplemented "EMBEDDED PDV"
 decodePrimitive (ASN1Header Universal 0xc _ _) p   = getUTF8String p
-decodePrimitive (ASN1Header Universal 0xd _ _) _   = Left $ ASN1NotImplemented "RELATIVE-OID"
+decodePrimitive (ASN1Header Universal 0xd _ _) _   = Left $ TypeNotImplemented "RELATIVE-OID"
 decodePrimitive (ASN1Header Universal 0x10 _ _) _  = error "sequence not a primitive"
 decodePrimitive (ASN1Header Universal 0x11 _ _) _  = error "set not a primitive"
 decodePrimitive (ASN1Header Universal 0x12 _ _) p  = getNumericString p
@@ -227,89 +231,90 @@ decodePrimitive (ASN1Header Universal 0x1e _ _) p  = getBMPString p
 decodePrimitive (ASN1Header tc        tag  _ _) p  = Right $ Other tc tag p
 
 
-getBoolean :: Bool -> ByteString -> Either ASN1Err ASN1
+getBoolean :: Bool -> ByteString -> Either ASN1Error ASN1
 getBoolean isDer s =
 	if B.length s == 1
-		then
-			case B.head s of
-				0    -> Right (Boolean False)
-				0xff -> Right (Boolean True)
-				_    -> if isDer then Left $ ASN1PolicyFailed "DER" "boolean value not canonical" else Right (Boolean True)
-		else Left $ ASN1Misc "boolean: length not within bound"
+		then case B.head s of
+			0    -> Right (Boolean False)
+			0xff -> Right (Boolean True)
+			_    -> if isDer then Left $ PolicyFailed "DER" "boolean value not canonical" else Right (Boolean True)
+		else Left $ TypeDecodingFailed "boolean: length not within bound"
 
 {- | getInteger, parse a value bytestring and get the integer out of the two complement encoded bytes -}
-getInteger :: ByteString -> Either ASN1Err ASN1
+getInteger :: ByteString -> Either ASN1Error ASN1
 getInteger s
-	| B.length s == 0 = Left $ ASN1Misc "integer: null encoding"
+	| B.length s == 0 = Left $ TypeDecodingFailed "integer: null encoding"
 	| B.length s == 1 = Right $ IntVal $ snd $ intOfBytes s
 	| otherwise       =
 		if (v1 == 0xff && testBit v2 7) || (v1 == 0x0 && (not $ testBit v2 7))
-			then Left $ ASN1Misc "integer: not shortest encoding"
+			then Left $ TypeDecodingFailed "integer: not shortest encoding"
 			else Right $ IntVal $ snd $ intOfBytes s
 		where
 			v1 = s `B.index` 0
 			v2 = s `B.index` 1
 
 
-getBitString :: ByteString -> Either ASN1Err ASN1
+getBitString :: ByteString -> Either ASN1Error ASN1
 getBitString s =
 	let toSkip = B.head s in
 	let toSkip' = if toSkip >= 48 && toSkip <= 48 + 7 then toSkip - (fromIntegral $ ord '0') else toSkip in
 	let xs = B.tail s in
 	if toSkip' >= 0 && toSkip' <= 7
 		then Right $ BitString $ toBitArray (L.fromChunks [xs]) (fromIntegral toSkip')
-		else Left $ ASN1Misc ("bitstring: skip number not within bound " ++ show toSkip' ++ " " ++  show s)
+		else Left $ TypeDecodingFailed ("bitstring: skip number not within bound " ++ show toSkip' ++ " " ++  show s)
 
-getString :: (ByteString -> Maybe ASN1Err) -> ByteString -> Either ASN1Err L.ByteString
+getString :: (ByteString -> Maybe ASN1Error) -> ByteString -> Either ASN1Error L.ByteString
 getString check s =
 	case check s of
 		Nothing  -> Right $ L.fromChunks [s]
 		Just err -> Left err
 
-getOctetString :: ByteString -> Either ASN1Err ASN1
-getOctetString = either Left (Right . OctetString) . getString (\_ -> Nothing)
+getOctetString :: ByteString -> Either ASN1Error ASN1
+getOctetString = (OctetString <$>) . getString (\_ -> Nothing)
 
-getNumericString :: ByteString -> Either ASN1Err ASN1
-getNumericString = either Left (Right . NumericString) . getString (\_ -> Nothing)
+getNumericString :: ByteString -> Either ASN1Error ASN1
+getNumericString = (NumericString <$>) . getString (\_ -> Nothing)
 
-getPrintableString :: ByteString -> Either ASN1Err ASN1
-getPrintableString = either Left (Right . PrintableString . T.unpack . decodeASCII) . getString (\_ -> Nothing)
+getPrintableString :: ByteString -> Either ASN1Error ASN1
+getPrintableString = (PrintableString . T.unpack . decodeASCII <$>) . getString (\_ -> Nothing)
 
-getUTF8String :: ByteString -> Either ASN1Err ASN1
-getUTF8String = either Left (Right . UTF8String . T.unpack . decodeUtf8) . getString (\_ -> Nothing)
+getUTF8String :: ByteString -> Either ASN1Error ASN1
+getUTF8String = (UTF8String . T.unpack . decodeUtf8 <$>) . getString (\_ -> Nothing)
 
-getT61String :: ByteString -> Either ASN1Err ASN1
-getT61String = either Left (Right . T61String . T.unpack . decodeASCII) . getString (\_ -> Nothing)
+getT61String :: ByteString -> Either ASN1Error ASN1
+getT61String = (T61String . T.unpack . decodeASCII <$>) . getString (\_ -> Nothing)
 
-getVideoTexString :: ByteString -> Either ASN1Err ASN1
-getVideoTexString = either Left (Right . VideoTexString) . getString (\_ -> Nothing)
+getVideoTexString :: ByteString -> Either ASN1Error ASN1
+getVideoTexString = (VideoTexString <$>) . getString (\_ -> Nothing)
 
-getIA5String :: ByteString -> Either ASN1Err ASN1
-getIA5String = either Left (Right . IA5String . T.unpack . decodeASCII) . getString (\_ -> Nothing)
+getIA5String :: ByteString -> Either ASN1Error ASN1
+getIA5String = (IA5String . T.unpack . decodeASCII <$>) . getString (\_ -> Nothing)
 
-getGraphicString :: ByteString -> Either ASN1Err ASN1
-getGraphicString = either Left (Right . GraphicString) . getString (\_ -> Nothing)
+getGraphicString :: ByteString -> Either ASN1Error ASN1
+getGraphicString = (GraphicString <$>) . getString (\_ -> Nothing)
 
-getVisibleString :: ByteString -> Either ASN1Err ASN1
-getVisibleString = either Left (Right . VisibleString) . getString (\_ -> Nothing)
+getVisibleString :: ByteString -> Either ASN1Error ASN1
+getVisibleString = (VisibleString <$>) . getString (\_ -> Nothing)
 
-getGeneralString :: ByteString -> Either ASN1Err ASN1
-getGeneralString = either Left (Right . GeneralString) . getString (\_ -> Nothing)
+getGeneralString :: ByteString -> Either ASN1Error ASN1
+getGeneralString = (GeneralString <$>) . getString (\_ -> Nothing)
 
-getUniversalString :: ByteString -> Either ASN1Err ASN1
-getUniversalString = either Left (Right . UniversalString . T.unpack . decodeUtf32BE) . getString (\_ -> Nothing)
+getUniversalString :: ByteString -> Either ASN1Error ASN1
+getUniversalString = (UniversalString . T.unpack . decodeUtf32BE <$>) . getString (\_ -> Nothing)
 
-getCharacterString :: ByteString -> Either ASN1Err ASN1
-getCharacterString = either Left (Right . CharacterString) . getString (\_ -> Nothing)
+getCharacterString :: ByteString -> Either ASN1Error ASN1
+getCharacterString = (CharacterString <$>) . getString (\_ -> Nothing)
 
-getBMPString :: ByteString -> Either ASN1Err ASN1
-getBMPString = either Left (Right . BMPString . T.unpack . decodeUCS2BE) . getString (\_ -> Nothing)
+getBMPString :: ByteString -> Either ASN1Error ASN1
+getBMPString = (BMPString . T.unpack . decodeUCS2BE <$>) . getString (\_ -> Nothing)
 
-getNull :: ByteString -> Either ASN1Err ASN1
-getNull s = if B.length s == 0 then Right Null else Left $ ASN1Misc "Null: data length not within bound"
+getNull :: ByteString -> Either ASN1Error ASN1
+getNull s
+    | B.length s == 0 = Right Null
+    | otherwise       = Left $ TypeDecodingFailed "Null: data length not within bound"
 
 {- | return an OID -}
-getOID :: ByteString -> Either ASN1Err ASN1
+getOID :: ByteString -> Either ASN1Error ASN1
 getOID s = Right $ OID $ (fromIntegral (x `div` 40) : fromIntegral (x `mod` 40) : groupOID xs)
 	where
 		(x:xs) = B.unpack s
@@ -327,7 +332,7 @@ getOID s = Right $ OID $ (fromIntegral (x `div` 40) : fromIntegral (x `mod` 40)
 		spanSubOIDbound (a:as) = if testBit a 7 then (clearBit a 7 : ys, zs) else ([a], as)
 			where (ys, zs) = spanSubOIDbound as
 
-getUTCTime :: ByteString -> Either ASN1Err ASN1
+getUTCTime :: ByteString -> Either ASN1Error ASN1
 getUTCTime s =
 	case B.unpack s of
 		[y1, y2, m1, m2, d1, d2, h1, h2, mi1, mi2, s1, s2, z] ->
@@ -339,11 +344,11 @@ getUTCTime s =
 			let minute = integerise mi1 mi2 in
 			let second = integerise s1 s2 in
 			Right $ UTCTime (year, month, day, hour, minute, second, z == 90)
-		_                                                     -> Left $ ASN1Misc "utctime unexpected format"
+		_                                                     -> Left $ TypeDecodingFailed "utctime unexpected format"
 	where
 		integerise a b = ((fromIntegral a) - (ord '0')) * 10 + ((fromIntegral b) - (ord '0'))
 
-getGeneralizedTime :: ByteString -> Either ASN1Err ASN1
+getGeneralizedTime :: ByteString -> Either ASN1Error ASN1
 getGeneralizedTime s =
 	case B.unpack s of
 		[y1, y2, y3, y4, m1, m2, d1, d2, h1, h2, mi1, mi2, s1, s2, z] ->
@@ -354,7 +359,7 @@ getGeneralizedTime s =
 			let minute = integerise mi1 mi2 in
 			let second = integerise s1 s2 in
 			Right $ GeneralizedTime (year, month, day, hour, minute, second, z == 90)
-		_                                                     -> Left $ ASN1Misc "utctime unexpected format"
+		_                                                     -> Left $ TypeDecodingFailed "utctime unexpected format"
 	where
 		integerise a b = ((fromIntegral a) - (ord '0')) * 10 + ((fromIntegral b) - (ord '0'))
 
